package com.dayorganiser.jgantt.drawing.component;

import com.dayorganiser.jgantt.model.Axis;

import com.dayorganiser.jgantt.drawing.Drawing;
import com.dayorganiser.jgantt.drawing.Plotter;
import com.dayorganiser.jgantt.drawing.Proxy;

import com.dayorganiser.jgantt.drawing.agent.DrawingHelperAgent;

import com.dayorganiser.jgantt.drawing.drawing.DrawingConstants;
import com.dayorganiser.jgantt.drawing.drawing.DrawingHelper;

import com.dayorganiser.jgantt.model.DrawingModel;

import com.dayorganiser.jgantt.model.axis.AxisListener;

import com.dayorganiser.jgantt.model.event.ModelListener;

import com.dayorganiser.util.Interval;
import com.dayorganiser.util.Trace;

import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;

import java.awt.event.MouseListener;

import java.util.Collection;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.Map;

import javax.swing.JComponent;

/** the diagram panel is the heart of the histogram */
public class DrawingComponent extends JComponent
	implements Drawing
{
	private static String THIS = DrawingComponent.class.getName();
	private static boolean TRACE = Trace.getEnable(THIS);

	protected int minEntryWidth = 2;

	protected DrawingModel model;

	protected Map entryKeys; // entryKey : rectangle

	// these are used for rendering
	protected Map axises = new HashMap(2);
	protected Map plotters = new HashMap(5);
	protected Map proxies = new HashMap(5);

	protected DrawingHelper drawingHelper;

	protected Set axisListeners = new HashSet(); // this is the set of Axies we are listening to

	protected AxisListener axisListener = new LocalAxisListener();
	protected ModelListener modelListener = new LocalModelListener();

	public DrawingComponent()
	{
		this.drawingHelper = createDrawingHelper();
	}

	//___________________________________________________________________________

	public void initialise()
	{
	}

	public void terminate()
	{
		clearAxisListeners();
		axisListener = null;

		if (model != null)
			model.removeModelListener(modelListener);

		modelListener = null;
	}

	//___________________________________________________________________________

	public DrawingModel getDrawingModel()
	{
		return model;
	}

	public void setDrawingModel(DrawingModel drawingModel)
	{
		if (this.model != null)
		   this.model.removeModelListener(modelListener);

		drawingModel.addModelListener(modelListener);
		this.model = drawingModel;
	}

	//___________________________________________________________________________

	public void addAxis(Object key, Axis axis)
	{
		axises.put(key, axis);
	}

	public void addPlotter(Object key, Plotter plotter)
	{
		plotters.put(key, plotter);
	}

	public void addProxy(Object key, Proxy proxy)
	{
		proxies.put(key, proxy);
	}

	//___________________________________________________________________________

	public Axis getAxis(Object key)
	{
		return (Axis) axises.get(key);
	}

	/** @deprecated : should this be in the proxy */
	public void setMinimumEntryWidth(int minEntryWidth)
	{
		this.minEntryWidth = minEntryWidth;
	}

	//___________________________________________________________________________

	/** this can only return you an entry key */
	public Object getKeyAt(Point point, boolean returnParent)
	{
		for (Iterator viter = entryKeys.entrySet().iterator(); viter.hasNext();)
		{
			Map.Entry entryMap = (Map.Entry) viter.next();
		   Rectangle entryBounds = (Rectangle) entryMap.getValue();
		   if (entryBounds.contains(point))
				return entryMap.getKey();
		}
		return null;
	}

	//___________________________________________________________________________

	public void paintComponent(Graphics g)
	{
		super.paintComponent(g);
		try
		{
			if (getWidth() > 0 && getHeight() > 0)	// paints the diagrams along with there underlying entries
				paintEntries(g, model.getEntries());
		}
		catch (Exception e)
		{
		   e.printStackTrace(Trace.out);
		}
	}

	//___________________________________________________________________________

	public void addMouseListener(MouseListener listener)
	{
		if (listener instanceof DrawingHelperAgent)
		   ((DrawingHelperAgent) listener).setDrawingHelper(drawingHelper);

		super.addMouseListener(listener);
	}

	public void removeMouseListener(MouseListener listener)
	{
		super.removeMouseListener(listener);
	}

	//___________________________________________________________________________

	protected void addAxisListener(Axis axis)
	{
		axisListeners.add(axis);
		axis.addAxisListener(axisListener);
	}

	protected void removeAxisListener(Axis axis)
	{
		axis.removeAxisListener(axisListener);
		axisListeners.remove(axis);
	}

	//___________________________________________________________________________

	protected void clearAxisListeners()
	{
		for (Iterator iter = axisListeners.iterator(); iter.hasNext();)
		{
		   Axis axis = (Axis) iter.next();
			axis.removeAxisListener(axisListener);
		   iter.remove();
		}
	}

	//___________________________________________________________________________

	protected DrawingHelper createDrawingHelper()
	{
		return new LocalDrawingHelper();
	}

	//___________________________________________________________________________

	protected void paintEntries(Graphics g, Collection values)
	{
		//int timeWidth = g.getClipBounds().width; // timeWidth should be declared obsolete
	   clearAxisListeners();

   	// setup an entry in the entry map
		Map keyValues = new HashMap(values.size());

	   Rectangle bounds = g.getClipBounds();
		Rectangle offset = new Rectangle(0,0,0,0);
		for (Iterator iter = values.iterator(); iter.hasNext();)
		{
		   Object entryKey = iter.next();

		   Axis xAxis = (Axis) axises.get(model.getAxis(entryKey, DrawingConstants.HORIZONTAL_AXIS));
		   Axis yAxis = (Axis) axises.get(model.getAxis(entryKey, DrawingConstants.VERTICAL_AXIS));

		   // now that we have the axis's we need to listen to changes with in them
   	   addAxisListener(xAxis);
		   addAxisListener(yAxis);

		   Interval xInterval = model.getInterval(entryKey, DrawingConstants.HORIZONTAL_AXIS);
		   Interval yInterval = model.getInterval(entryKey, DrawingConstants.VERTICAL_AXIS);

 		   // don't waste our time painting entries that are not visible
		   if (isVisible(bounds.width, xAxis, xInterval) &&
				isVisible(bounds.height, yAxis, yInterval))
		   {
				Plotter plotter = (Plotter)
				   plotters.get(model.getPlotter(entryKey));
				if (plotter == null)
				   continue;

				Proxy proxy = (Proxy)
					proxies.get(model.getProxy(entryKey));
				if (proxy == null)
				   continue;

				// calculate the visible intervals
				int vX = intersects(xAxis, xInterval);
				int vY = intersects(yAxis, yInterval);

				// paint the component
				Rectangle rect = plotter.paintComponent(entryKey,
					createGraphics(g, xAxis, xInterval, yAxis, yInterval),
	   			   vX, vY, proxy, model.getAttributes(entryKey));

				// cache the used size (only necessary for selectable things
				if (rect != null)
				{
				   rect = new Rectangle( offset.x + rect.x, offset.y + rect.y,
						rect.width, rect.height);
					keyValues.put(entryKey, rect );
				}
		   }
		}
	}

	//___________________________________________________________________________

	protected Graphics createGraphics(Graphics g, Axis xAxis, Interval xInterval,
		Axis yAxis, Interval yInterval)
	{
		Rectangle bound = g.getClipBounds();
		int x1 = Math.max(xAxis.getPoint(bound.width, xInterval.getStart()), 0);
		int x2 = Math.min(xAxis.getPoint(bound.width, xInterval.getFinish()), bound.width);

		int y1 = Math.max(yAxis.getPoint(bound.height, yInterval.getStart()), 0);
		int y2 = Math.min(yAxis.getPoint(bound.height, yInterval.getFinish()), bound.height);

		return g.create(x1, y1, x2 - x1, y2 - y1);
	}

	//___________________________________________________________________________

	protected int intersects(Axis axis, Interval entry)
	{
		Interval visible = axis.getVisible();
		if (!visible.intersects(entry))
		   return Integer.MIN_VALUE;  // throw an exception

		int result = visible.containsValue(entry.getStart()) ? 1 : 0;
		result += visible.containsValue(entry.getFinish()) ? 2 : 0;
		return result;
	}

	/** returns a quick true on intersection */
	protected boolean isVisible(int width, Axis axis, Interval entry)
	{
		int start = axis.getPoint(width, entry.getStart());
		int finish = axis.getPoint(width, entry.getFinish());

		int entryWidth =  finish - Math.max(start,0);
	   entryWidth = entryWidth > width ? width : Math.max(this.minEntryWidth, entryWidth);
		return entryWidth > 0;
	}

	//___________________________________________________________________________

	protected class LocalDrawingHelper implements DrawingHelper
	{
		public Object getKeyAt(Point point, boolean returnParent)
		{
		   return DrawingComponent.this.getKeyAt(point,returnParent);
		}

		public void fireRepaint()
		{
		   DrawingComponent.this.repaint();
		}
	}

	//___________________________________________________________________________

	/** listens to n acceses */
	private class LocalAxisListener implements AxisListener
	{
		public void intervalChanged()
		{
		   repaint();
		}

		public void granularityChanged()
		{
		   repaint();
		}
	}

	//___________________________________________________________________________

	/** listens to n models */
	private class LocalModelListener implements ModelListener
	{
		public void refresh()
		{
		   repaint();
		}
	}
}